/*
 STORYBOOK: Open Source software for novelists and authors.
 Copyright (C) 2008 - 2012 Martin Mustun

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a fileCopy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package storybook.ui;

import i18n.I18N;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Frame;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JToolBar;
import javax.swing.UIManager;
import javax.swing.WindowConstants;
import javax.swing.border.EtchedBorder;
import net.infonode.docking.DockingWindow;
import net.infonode.docking.DockingWindowAdapter;
import net.infonode.docking.RootWindow;
import net.infonode.docking.SplitWindow;
import net.infonode.docking.TabWindow;
import net.infonode.docking.View;
import net.infonode.docking.ViewSerializer;
import net.infonode.docking.properties.RootWindowProperties;
import net.infonode.docking.theme.DockingWindowsTheme;
import net.infonode.docking.theme.ShapedGradientDockingTheme;
import net.infonode.docking.util.DockingUtil;
import net.infonode.docking.util.MixedViewHandler;
import net.infonode.docking.util.StringViewMap;
import net.infonode.util.Direction;
import org.hibernate.Session;
import org.miginfocom.swing.MigLayout;
import resources.icons.IconUtil;
import shef.ShefEditor;
import storybook.App;
import storybook.Const;
import storybook.Pref;
import storybook.action.ActionHandler;
import storybook.action.MainAction;
import storybook.action.SbActionManager;
import storybook.ctrl.Ctrl;
import storybook.edit.Editor;
import storybook.model.BlankModel;
import storybook.model.EntityUtil;
import storybook.model.H2File;
import storybook.model.Model;
import storybook.model.book.Book;
import storybook.model.hbn.dao.SceneDAOImpl;
import storybook.model.hbn.entity.AbstractEntity;
import storybook.model.hbn.entity.Chapter;
import storybook.model.hbn.entity.Part;
import storybook.model.hbn.entity.Scene;
import storybook.toolkit.DockUtil;
import storybook.toolkit.LOG;
import storybook.toolkit.SpellUtil;
import storybook.toolkit.calendar.SbCalendar;
import storybook.toolkit.file.EnvUtil;
import storybook.toolkit.file.IOUtil;
import storybook.toolkit.file.TempUtil;
import storybook.toolkit.swing.FontUtil;
import storybook.toolkit.swing.SwingUtil;
import storybook.ui.SbView.VIEWNAME;
import storybook.ui.interfaces.IPaintable;
import storybook.ui.panel.AbstractPanel;
import storybook.ui.panel.BlankPanel;
import storybook.ui.panel.typist.TypistPanel;
import storybook.ui.panel.typist.TypistScenario;

/**
 * @author martin
 *
 */
@SuppressWarnings("serial")
public class MainFrame extends JFrame implements IPaintable {

	private static final String TT = "MainFrame";

	public final static int SCREEN_WIDTH = 1000,
			SCREEN_HEIGHT = 700,
			SCREEN_X = 100,
			SCREEN_Y = 100;

	private Model model;
	private Ctrl ctrl;
	private SbActionManager sbActionManager;
	private SbViewFactory viewFactory;
	private JToolBar mainToolBar;
	private RootWindow rootWindow;
	public StatusBarPanel statusBar;
	@SuppressWarnings("MismatchedQueryAndUpdateOfCollection")
	private final HashMap<Integer, JComponent> dynamicViews = new HashMap<>();
	private H2File h2File;
	private boolean EditorModless;
	private Chapter lastChapter;
	private boolean updated = false;
	public boolean showAllParts = false;
	private TypistPanel typistScene;
	private TypistScenario typistScenario;
	private SbCalendar calendrier;
	public MainAction action;
	private boolean isTypist = false, isScenario = false;
	public JFrame fullFrame;
	private Book book;

	public MainFrame() {
		FontUtil.setDefaultFont(new Font("Arial", Font.PLAIN, 12));
	}

	public MainFrame(H2File dbF) {
		this();
		init(dbF);
		initUi();
	}

	@Override
	public void init() {
		//LOG.trace(TT+".init()");
		action = new MainAction(this);
		calendrier = new SbCalendar();
		h2File = null;
		viewFactory = new SbViewFactory(this);
		sbActionManager = new SbActionManager(this);
		sbActionManager.init();
		ctrl = new Ctrl(this);
		BlankModel blank = new BlankModel(this);
		ctrl.attachModel(blank);
		setIconImage(IconUtil.getIconImage("icon"));
		addWindowListener(new MainFrameWindowAdaptor());
	}

	public void init(H2File dbF) {
		//LOG.trace(TT+".init(" + dbF.getH2Name() + ")");
		LOG.setLogging();
		try {
			LOG.log("open file: " + dbF.getFile().getAbsolutePath());
			action = new MainAction(this);
			this.h2File = dbF;
			viewFactory = new SbViewFactory(this);
			viewFactory.setInitialisation();
			sbActionManager = new SbActionManager(this);
			sbActionManager.init();
			ctrl = new Ctrl(this);
			model = new Model(this);
			if (!dbF.getH2Name().isEmpty()) {
				model.initSession(dbF.getH2Name());
			}
			book = new Book(this);
			calendrier = Book.getCalendar(this);
			ctrl.attachModel(model);
			addWindowListener(new MainFrameWindowAdaptor());
			SpellUtil.registerDictionaries();
			viewFactory.resetInitialisation();
			setSummaryImage();
		} catch (Exception e) {
			LOG.err("error MainFrame.init(dbfname=" + dbF.getName() + ")", e);
		}
	}

	private void setSummaryImage() {
		// set base summary image
		String img = getImageDir();
		File f = new File(img);
		if (!f.exists()) {
			if (!f.mkdir()) {
				LOG.err(TT + " init error: unable to create \"" + img + "\"");
				return;
			}
		}
		img = img + File.separator + "summary.png";
		f = new File(img);
		if (!f.exists()) {
			try {
				//LOG.trace(TT + " copy summary image to \"" + img + "\"");
				IOUtil.resourceCopyTo("icons/png/summary.png", f);
			} catch (Exception ex) {
				LOG.err("unable to copy summary image to \"" + img + "\"", ex);
			}
		}
	}

	@Override
	public void initUi() {
		//LOG.trace(TT+".initUi()");
		ShefEditor.setSpelling(this.getPref().getString(Pref.KEY.SPELLING));
		setLayout(new MigLayout("flowy,fill,ins 0,gap 0", "", "[grow]"));
		setIconImage(IconUtil.getIconImage("icon"));
		App.getInstance().fontResetUi();
		setFont(App.getInstance().fontGetDefault());
		UIManager.put("default", App.getInstance().fontGetDefault());
		initRootWindow();
		setDefaultLayout();
		add(rootWindow, "grow");
		statusBar = new StatusBarPanel(this);
		add(statusBar, "growx");
		ctrl.attachView(statusBar);

		setTitle();
		restoreDimension();
		setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);

		pack();
		setVisible(true);
		initAfterPack();

		JMenuBar menubar = getJMenuBar();
		menubar.setFont(App.getInstance().fontGetDefault());
		ctrl.detachView(menubar);
		ctrl.attachView(menubar);
		// load last used layout
		DockUtil.layoutLoad(this, Pref.KEY.LAYOUT_LAST_USED.toString());
		typistSet();
		//LOG.trace("MainFrame end initUi");
	}

	public void typistActivate() {
		//LOG.trace(TT+".activateTipist()=>" + (isTypist ? "true" : "false"));
		if (isTypist) {
			if (book.info.scenarioGet() || isScenario) {
				typistScenario.closeScene();
				fullFrame.remove(typistScenario);
				isScenario = false;
			} else {
				typistScene.closeScene();
				fullFrame.remove(typistScene);
			}
			fullFrame.dispose();
			this.setVisible(true);
			isTypist = false;
			App.getInstance().preferences.setBoolean(Pref.KEY.TYPIST_USE, false);
		} else {
			this.setVisible(false);
			fullFrame = new JFrame();
			fullFrame.setIconImage(IconUtil.getIconImage("icon"));
			fullFrame.setLayout(new MigLayout("fill"));
			if (book.info.scenarioGet()) {
				typistScenario = new TypistScenario(this, lastUsedScene);
				fullFrame.add(typistScenario, "grow");
				isScenario = true;
			} else {
				typistScene = new TypistPanel(this, lastUsedScene);
				fullFrame.add(typistScene, "grow");
			}
			fullFrame.setUndecorated(true);
			fullFrame.setVisible(true);
			fullFrame.pack();
			fullFrame.setExtendedState(this.getExtendedState() | JFrame.MAXIMIZED_BOTH);
			fullFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
			isTypist = true;
			App.getInstance().preferences.setBoolean(Pref.KEY.TYPIST_USE, true);
		}
	}

	public void typistSet() {
		if (h2File.isOK()
				&& App.getInstance().preferences.getBoolean(Pref.KEY.TYPIST_USE)) {
			typistActivate();
		}
	}

	public void setTitle() {
		//LOG.trace(TT+".setTitle()");
		if (h2File != null) {
			String title = (isUpdated() ? "*" : "") + book.getTitle() + (isUpdated() ? "*" : "");
			setTitle(title + " - " + Const.getFullName());
		} else {
			setTitle(Const.getFullName());
		}
		if (sbActionManager != null && sbActionManager.getMainMenu() != null) {
			this.sbActionManager.getMainMenu().enableSave(isUpdated());
		}
		getSbActionManager().getMainMenu().setLinks();
	}

	private void initRootWindow() {
		//LOG.trace(TT+".initRootWindow()");
		StringViewMap viewMap = viewFactory.getViewMap();
		MixedViewHandler handler = new MixedViewHandler(viewMap, new ViewSerializer() {
			@Override
			public void writeView(View view, ObjectOutputStream out) throws IOException {
				out.writeInt(((DynamicView) view).getId());
			}

			@Override
			public View readView(ObjectInputStream in) throws IOException {
				return getDynamicView(in.readInt());
			}
		});
		rootWindow = DockingUtil.createRootWindow(viewMap, handler, true);
		rootWindow.setName("rootWindow");
		rootWindow.setPreferredSize(new Dimension(4096, 2048));
		// set theme
		DockingWindowsTheme currentTheme = new ShapedGradientDockingTheme();
		RootWindowProperties properties = new RootWindowProperties();
		properties.addSuperObject(currentTheme.getRootWindowProperties());
		// Our properties object is the super object of the root window
		// properties object, so all property values of the
		// theme and in our property object will be used by the root window
		rootWindow.getRootWindowProperties().addSuperObject(properties);
		rootWindow.setBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED));
	}

	public void setDefaultLayout() {
		//LOG.trace(TT+".setDefaultLayout()");
		SbView attributesView = getView(VIEWNAME.ATTRIBUTES);
		SbView categoriesView = getView(VIEWNAME.CATEGORIES);
		SbView chaptersView = getView(VIEWNAME.CHAPTERS);
		SbView endnotesView = getView(VIEWNAME.ENDNOTES);
		SbView episodesView = getView(VIEWNAME.EPISODES);
		SbView eventsView = getView(VIEWNAME.EVENTS);
		SbView gendersView = getView(VIEWNAME.GENDERS);
		SbView ideasView = getView(VIEWNAME.IDEAS);
		SbView internalsView = getView(VIEWNAME.INTERNALS);
		SbView itemsView = getView(VIEWNAME.ITEMS);
		SbView itemLinksView = getView(VIEWNAME.ITEMLINKS);
		SbView locationsView = getView(VIEWNAME.LOCATIONS);
		SbView memosView = getView(VIEWNAME.MEMOS);
		SbView partsView = getView(VIEWNAME.PARTS);
		SbView personsView = getView(VIEWNAME.PERSONS);
		SbView plotView = getView(VIEWNAME.PLOTS);
		SbView relationsView = getView(VIEWNAME.RELATIONS);
		SbView scenesView = getView(VIEWNAME.SCENES);
		SbView strandsView = getView(VIEWNAME.STRANDS);
		SbView tagsView = getView(VIEWNAME.TAGS);
		SbView tagLinksView = getView(VIEWNAME.TAGLINKS);

		SbView attributesListView = getView(VIEWNAME.ATTRIBUTESLIST);
		SbView bookView = getView(VIEWNAME.BOOK);
		SbView chronoView = getView(VIEWNAME.CHRONO);
		SbView timelineView = getView(VIEWNAME.TIMELINE);
		SbView infoView = getView(VIEWNAME.INFO);
		SbView manageView = getView(VIEWNAME.MANAGE);
		SbView memoriaView = getView(VIEWNAME.MEMORIA);
		SbView navigationView = getView(VIEWNAME.NAVIGATION);
		SbView planningView = getView(VIEWNAME.PLAN);
		SbView readingView = getView(VIEWNAME.READING);
		SbView scenarioView = getView(VIEWNAME.SCENARIO);
		SbView storyboard = getView(VIEWNAME.STORYBOARD);
		SbView storymap = getView(VIEWNAME.STORYMAP);
		SbView treeView = getView(VIEWNAME.TREE);
		SbView typistView = getView(VIEWNAME.TYPIST);

		SbView chartPersonsByDate = getView(VIEWNAME.CHART_PERSONS_BY_DATE);
		SbView chartPersonsByScene = getView(VIEWNAME.CHART_PERSONS_BY_SCENE);
		SbView chartWiWW = getView(VIEWNAME.CHART_WIWW);
		SbView chartStrandsByDate = getView(VIEWNAME.CHART_STRANDS_BY_DATE);
		SbView chartOccurrenceOfPersons = getView(VIEWNAME.CHART_OCCURRENCE_OF_PERSONS);
		SbView chartOccurrenceOfLocations = getView(VIEWNAME.CHART_OCCURRENCE_OF_LOCATIONS);
		SbView chartOccurrenceOfItems = getView(VIEWNAME.CHART_OCCURRENCE_OF_ITEMS);
		SbView chartGantt = getView(VIEWNAME.CHART_GANTT);

		TabWindow tabInfoNavi = new TabWindow(new SbView[]{infoView, navigationView});
		tabInfoNavi.setName("tabInfoNaviWindow");
		SplitWindow swTreeInfo = new SplitWindow(false, 0.6f, treeView, tabInfoNavi);
		swTreeInfo.setName("swTreeInfo");
		TabWindow tabWindow;
		tabWindow = new TabWindow(new SbView[]{
			//les vues
			attributesListView,
			chronoView,
			timelineView,
			bookView,
			manageView,
			planningView,
			readingView,
			memoriaView,
			scenarioView,
			storyboard,
			storymap,
			typistView,
			//les principaux
			partsView,
			chaptersView,
			scenesView,
			locationsView,
			personsView,
			itemsView,
			//les autres
			attributesView,
			categoriesView,
			endnotesView,
			episodesView,
			eventsView,
			gendersView,
			ideasView,
			internalsView,
			itemLinksView,
			memosView,
			relationsView,
			plotView,
			strandsView,
			tagsView,
			tagLinksView,
			//les charts
			chartPersonsByDate,
			chartPersonsByScene,
			chartWiWW,
			chartStrandsByDate,
			chartOccurrenceOfPersons,
			chartOccurrenceOfItems,
			chartOccurrenceOfLocations,
			chartGantt
		});
		tabWindow.setName("tabWindow");
		SplitWindow swMain = new SplitWindow(true, 0.20f, swTreeInfo, tabWindow);
		swMain.setName("swMain");
		rootWindow.setWindow(swMain);
		//fermeture de toutes les vues sauf bookView
		attributesListView.close();
		//bookView.close();
		chronoView.close();
		timelineView.close();
		infoView.restoreFocus();
		manageView.close();
		memoriaView.close();
		planningView.close();
		readingView.close();
		scenarioView.close();
		storyboard.close();
		storymap.close();
		typistView.close();
		//fermeture de toutes les tables sauf scenes, lieux, personnages
		attributesView.close();
		categoriesView.close();
		chaptersView.close();
		endnotesView.close();
		episodesView.close();
		eventsView.close();
		gendersView.close();
		ideasView.close();
		internalsView.close();
		itemsView.close();
		itemLinksView.close();
		//locationsView.close();
		memosView.close();
		partsView.close();
		//personsView.close();
		plotView.close();
		relationsView.close();
		//scenesView.close();
		strandsView.close();
		tagsView.close();
		tagLinksView.close();
		//fermeture des charts
		chartPersonsByDate.close();
		chartPersonsByScene.close();
		chartWiWW.close();
		chartStrandsByDate.close();
		chartOccurrenceOfPersons.close();
		chartOccurrenceOfItems.close();
		chartOccurrenceOfLocations.close();
		chartGantt.close();
		//restauration du focus sur bookView
		bookView.restoreFocus();
		rootWindow.getWindowBar(Direction.RIGHT).setEnabled(true);
		DockUtil.setRespectMinimumSize(this);
	}

	private void initAfterPack() {
		SbView scenesView = getView(VIEWNAME.SCENES);
		SbView locationsView = getView(VIEWNAME.LOCATIONS);
		SbView personsView = getView(VIEWNAME.PERSONS);
		SbView bookView = getView(VIEWNAME.BOOK);
		SbView treeView = getView(VIEWNAME.TREE);
		SbView quickInfoView = getView(VIEWNAME.INFO);
		SbView navigationView = getView(VIEWNAME.NAVIGATION);
		// add docking window adapter to all views (except editor)
		MainDockingWindowAdapter dockingAdapter = new MainDockingWindowAdapter();
		for (int i = 0; i < viewFactory.getViewMap().getViewCount(); ++i) {
			View view = viewFactory.getViewMap().getViewAtIndex(i);
			view.addListener(dockingAdapter);
		}
		// load initially shown views here
		SbView[] views2 = {scenesView,
			personsView, locationsView,
			bookView, treeView, quickInfoView, navigationView};
		for (SbView view : views2) {
			if (!viewFactory.loadView(view)) {
				continue;
			}
			ctrl.attachView(view.getComponent());
			if (view.isLoaded()) {
				model.fireAgain(view);
			}
		}
		quickInfoView.restoreFocus();
		bookView.restoreFocus();
	}

	public SbView getView(AbstractEntity viewName) {
		return viewFactory.getView(viewName);
	}

	public SbView getView(String viewName) {
		return viewFactory.getView(viewName);
	}

	public SbView getView(VIEWNAME viewName) {
		return viewFactory.getView(viewName);
	}

	public void showView(VIEWNAME viewName) {
		//LOG.trace(TT+".showView(" + viewName.name() + ")");
		cursorSetWaiting();
		SbView view = getView(viewName);
		if (view.getRootWindow() != null) {
			view.restoreFocus();
		} else {
			//LOG.trace(">>> RootWindow=null");
			DockingUtil.addWindow(view, rootWindow);
		}
		view.requestFocusInWindow();
		DockUtil.setRespectMinimumSize(this);
		cursorSetDefault();
	}

	public void showAndFocus(VIEWNAME viewName) {
		//LOG.trace(TT+".showAndFocus(" + viewName.name() + ")");
		System.gc();
		SbView view = getView(viewName);
		view.restore();
		view.restoreFocus();
	}

	public void closeView(VIEWNAME viewName) {
		//LOG.trace(TT + ".closeView(" + viewName.name() + ")");
		SbView view = getView(viewName);
		view.close();
	}

	public void refresh() {
		//LOG.trace(TT+".refresh()");
		cursorSetWaiting();
		for (int i = 0; i < viewFactory.getViewMap().getViewCount(); ++i) {
			SbView view = (SbView) viewFactory.getViewMap().getViewAtIndex(i);
			viewFactory.setViewTitle(view);
			getBookController().refresh(view);
		}
		cursorSetDefault();
	}

	public void refreshStatusBar() {
		//LOG.trace(TT + ".refreshStatusBar()");
		if (statusBar != null) {
			statusBar.refresh();
		}
	}

	public void showEditor() {
		//LOG.trace(TT+".showEditor()");
		/*SwingUtilities.invokeLater(new Runnable() {
		 @Override
		 public void run() {
		 LOG.trace(TT+"showEditor()-->run");
		 SbView editorView = getView(VIEWNAME.EDITOR);
		 editorView.cleverRestoreFocus();
		 }
		 });*/
	}

	public void initBlankUi() {
		h2File = null;
		setTitle(Const.getFullName());
		Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
		setLocation(screen.width / 2 - 450, screen.height / 2 - 320);
		setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
		App.getInstance().fontResetUi();
		sbActionManager.reloadMenuToolbar();
		BlankPanel blankPanel = new BlankPanel(this);
		blankPanel.initAll();
		add(blankPanel);
		pack();
		setVisible(true);
	}

	public void cursorSetDefault() {
		SwingUtil.setDefaultCursor(this);
	}

	public void cursorSetWaiting() {
		SwingUtil.setWaitingCursor(this);
	}

	public H2File getH2File() {
		return h2File;
	}

	public boolean isBlank() {
		return h2File == null;
	}

	public Ctrl getBookController() {
		return ctrl;
	}

	public Model getBookModel() {
		return model;
	}

	public Session getSession() {
		return (model.getSession());
	}

	public RootWindow getRootWindow() {
		return rootWindow;
	}

	public SbActionManager getSbActionManager() {
		return sbActionManager;
	}

	public ActionHandler getActionController() {
		return sbActionManager.getActionController();
	}

	public SbViewFactory getViewFactory() {
		return viewFactory;
	}

	private MainFrame getThis() {
		return this;
	}

	public boolean isMaximized() {
		return (getExtendedState() & Frame.MAXIMIZED_BOTH) == Frame.MAXIMIZED_BOTH;
	}

	public void setMaximized() {
		setExtendedState(Frame.MAXIMIZED_BOTH);
	}

	public void close(boolean forceExit) {
		//LOG.trace(TT+".close(forceExit=" + (forceExit ? "true" : "false") + ")");
		App app = App.getInstance();
		if (!isBlank()) {
			if (isTypist) {
				if (isScenario) {
					if (typistScenario.askModified() != JOptionPane.YES_OPTION) {
						return;
					}
					typistScenario.closeScene();
				} else {
					if (typistScene.askModified() != JOptionPane.YES_OPTION) {
						return;
					}
					typistScene.closeScene();
				}
			}
			if (isUpdated()) {
				//check if data are modified
				Object[] choix = {
					I18N.getMsg("cancel"),
					I18N.getMsg("file.save"),
					I18N.getMsg("ignore")
				};
				int n = JOptionPane.showOptionDialog(null,
						I18N.getMsg("close.confirm"),
						I18N.getMsg("close"),
						JOptionPane.YES_NO_OPTION,
						JOptionPane.QUESTION_MESSAGE,
						null, choix, choix[0]);
				if (n == 0) {
					return;
				}
				if (n == 1) {
					h2File.save();
				}
				if (getPref().getBoolean(Pref.KEY.CONFIRM_EXIT, true)) {
					int r = JOptionPane.showConfirmDialog(getThis(),
							I18N.getMsg("ask.close"),
							I18N.getMsg("close"),
							JOptionPane.YES_NO_OPTION);
					if (r == JOptionPane.NO_OPTION || n == JOptionPane.CLOSED_OPTION) {
						return;
					}
				}
			}
			viewFactory.saveAllTableDesign();
			saveContext();
			app.updateFilePref(this.getH2File(), book.getTitle());
		} else if (this.getH2File() != null) {
			saveContext();
			app.updateFilePref(this.getH2File(), book.getTitle());
		}
		if (model != null) {
			model.closeSession();
		}
		app.removeMainFrame(this);
		dispose();
		if (app.getMainFrames().isEmpty()) {
			if (forceExit) {
				app.exit();
			} else { //re-create blank
				MainFrame mainFrame = new MainFrame();
				mainFrame.init();
				mainFrame.initBlankUi();
				app.addMainFrame(mainFrame);
			}
		}
	}

	private View getDynamicView(int id) {
		View view = (View) dynamicViews.get(id);
		if (view == null) {
			view = new DynamicView("Dynamic View " + id, null, createDummyViewComponent("Dynamic View " + id), id);
		}
		return view;
	}

	private static JComponent createDummyViewComponent(String text) {
		StringBuilder sb = new StringBuilder();
		for (int j = 0; j < 100; j++) {
			sb.append(text).append(". This is line ").append(j).append("\n");
		}
		return new JScrollPane(new JTextArea(sb.toString()));
	}

	private void restoreDimension() {
		int w = getPref().getInteger(Pref.KEY.WINDOW_SIZEW, SCREEN_WIDTH);
		int h = getPref().getInteger(Pref.KEY.WINDOW_SIZEH, SCREEN_HEIGHT);
		setPreferredSize(new Dimension(w, h));
		int x = getPref().getInteger(Pref.KEY.WINDOW_POSX, SCREEN_X);
		int y = getPref().getInteger(Pref.KEY.WINDOW_POSY, SCREEN_Y);

		// Get screens rectangle
		Rectangle virtualBounds = new Rectangle();
		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsDevice[] gs = ge.getScreenDevices();
		for (GraphicsDevice device : gs) {
			virtualBounds.add(device.getDefaultConfiguration().getBounds());
		}

		// Do not put frame out of screens
		x = (int) Math.max(virtualBounds.getMinX(), Math.min(virtualBounds.getMaxX() - w, x));
		y = (int) Math.max(virtualBounds.getMinY(), Math.min(virtualBounds.getMaxY() - h, y));
		setLocation(x, y);

		boolean maximized = getPref().getBoolean(Pref.KEY.WINDOW_MAXIMIZED, false);
		if (maximized) {
			setMaximized();
		}
	}

	public Pref getPref() {
		return (App.getPref());
	}

	public void chapterSetLast(Chapter c) {
		lastChapter = c;
	}

	public Chapter chapterGetLast() {
		return (lastChapter);
	}

	public void changePath(String oldPath, String newPath) {
		//LOG.trace(TT+".changePath(" + oldPath + ", " + newPath + ")");
		if (oldPath.equals(newPath)) {
			return;
		}
		Session session = model.beginTransaction();
		SceneDAOImpl SceneDAO = new SceneDAOImpl(session);
		List<Scene> scenes = SceneDAO.findAll();
		for (Scene scene : scenes) {
			String text = scene.getSummary();
			if (scene.getSummary().contains(oldPath)) {
				text = text.replace(oldPath, newPath);
				scene.setSummary(text);
				this.getBookController().updateEntity(scene);
			}
		}
		model.commit();
	}

	public void setUpdated() {
		//LOG.trace(TT+".setUpdated()");
		updated = true;
		setTitle();
	}

	public void setUpdated(boolean b) {
		//LOG.trace(TT+".setUpdated(b="+(b?"true":"false")+")");
		updated = b;
		setTitle();
	}

	public boolean isUpdated() {
		return updated;
	}

	private Scene lastUsedScene = null;

	public void sceneSetLast(Scene s) {
		lastUsedScene = s;
	}

	public Scene sceneGetLast() {
		if (lastUsedScene == null) {
			return ((Scene) EntityUtil.findEntities(this, Book.TYPE.SCENE).get(0));
		}
		return (lastUsedScene);
	}

	public String getImageDir() {
		/*String dir = Book.getString(this, "ImageDir");
		if (dir == null || dir.isEmpty()) {
			dir = h2File.getPath() + File.separator + "Images";
		}
		return dir;*/
		return h2File.getPath() + File.separator + "Images";
	}

	public String getPhotoDir() {
		String dir = getPref().galleryGetFolder();
		if (dir == null || dir.isEmpty()) {
			dir = h2File.getPath();
		}
		return (dir);
	}

	public void setImageDir(String str) {
		Book.setString(this, "ImageDir", str);
	}

	public void saveContext() {
		// save dimension, location, maximized
		if (this.isVisible()) {
			Dimension dim = this.getSize();
			dim.width += 10;//because MigLayout reduce this by 10
			getPref().setInteger(Pref.KEY.WINDOW_SIZEW, dim.width);
			getPref().setInteger(Pref.KEY.WINDOW_SIZEH, dim.height);
			getPref().setInteger(Pref.KEY.WINDOW_POSX, getLocationOnScreen().x);
			getPref().setInteger(Pref.KEY.WINDOW_POSY, getLocationOnScreen().y);
			getPref().setBoolean(Pref.KEY.WINDOW_MAXIMIZED, isMaximized());
			String s = "North";
			if (mainToolBar.getOrientation() == 0) {
				if (mainToolBar.getY() != 0) {
					s = "South";
				}
			} else if (mainToolBar.getX() != 0) {
				s = "East";
			} else {
				s = "West";
			}
			getPref().setString(Pref.KEY.TOOLBAR_ORIENTATION, s);
			DockUtil.layoutSave(this, Pref.KEY.LAYOUT_LAST_USED.toString());
			getPref().save();
		}
	}

	public void saveFile(boolean force) {
		//LOG.trace(TT+"saveFile()");
		if (!force && !isUpdated()) {
			return;
		}
		if (h2File != null) {
			viewFactory.saveAllTableDesign();
			Book.setBoolean(this, "TypistUse", isTypist);
			h2File.save();
			App.getInstance().updateFilePref(this.getH2File(), book.getTitle());
		}
		updated = false;
		setTitle();
	}

	public void reloadBook() {
		book = new Book(this);
		SbView view = getView(VIEWNAME.READING);
		if (view != null && view.isLoaded()) {
			getBookModel().fireAgain(view);
		}
	}

	public Book getBook() {
		return (book);
	}

	private Part lastPart = null;

	public Part getLastPart() {
		return lastPart;
	}

	public void setLastPart(Part part) {
		this.lastPart = part;
	}

	void pasteAsPart() {
		LOG.trace(TT + "pasteAsPart()");
		Part c = new Part();
		c.setDescription(getClipboardContents());
		showEditorAsDialog(c);
	}

	void pasteAsChapter() {
		LOG.trace(TT + "pasteAsChapter()");
		Chapter c = new Chapter();
		c.setDescription(getClipboardContents());
		showEditorAsDialog(c);
	}

	void pasteAsScene() {
		LOG.trace(TT + "pasteAsScene()");
		Scene c = new Scene();
		c.setSummary(getClipboardContents());
		showEditorAsDialog(c);
	}

	public void sceneSetLastDate(Date date) {
		this.getPref().setDate(Pref.KEY.SCENE_LASTDATE, date);
	}

	public Date sceneGetLastDate() {
		return this.getPref().getDate(Pref.KEY.SCENE_LASTDATE);
	}

	private static class DynamicView extends View {

		private final int id;

		DynamicView(String title, Icon icon, Component component, int id) {
			super(title, icon, component);
			this.id = id;
		}

		public int getId() {
			return id;
		}
	}

	private class MainFrameWindowAdaptor extends WindowAdapter {

		@Override
		public void windowClosing(WindowEvent evt) {
			close(true);
		}
	}

	private class MainDockingWindowAdapter extends DockingWindowAdapter {

		@Override
		@SuppressWarnings("null")
		public void windowAdded(DockingWindow addedToWindow, DockingWindow addedWindow) {
			//LOG.trace("MainDockingWindowAdapter.windowAdded(" + addedToWindow.getName() + ", " + addedWindow.getName() + ")");
			if (addedWindow instanceof SbView) {
				SbView view = (SbView) addedWindow;
				if (!view.isLoaded() && viewFactory.loadView(view)) {
					ctrl.attachView(view.getComponent());
					model.fireAgain(view);
				}
			}
		}

		@Override
		@SuppressWarnings("null")
		public void windowClosed(DockingWindow window) {
			//LOG.trace("MainDockingWindowAdapter.windowClosed(" + window.getName() + ")");
			if (window instanceof SbView) {
				SbView view = (SbView) window;
				if (!view.isLoaded()) {
					return;
				}
				ctrl.detachView((AbstractPanel) view.getComponent());
				viewFactory.unloadView(view);
			}
		}
	}

	public void setMainToolBar(JToolBar toolBar) {
		if (mainToolBar != null) {
			getContentPane().remove(mainToolBar);
		}
		this.mainToolBar = toolBar;
		String bl = this.getPref().getString(Pref.KEY.TOOLBAR_ORIENTATION, "North");
		if ("East".equals(bl) || "West".equals(bl)) {
			mainToolBar.setOrientation(1);
		}
		getContentPane().add(mainToolBar, bl);
	}

	public JToolBar getMainToolBar() {
		return mainToolBar;
	}

	/**
	 * call the dialog editor
	 *
	 * @param entity
	 * @param bt: optional, locate the dialog upon a button
	 *
	 * @return
	 */
	public boolean showEditorAsDialog(AbstractEntity entity, JButton... bt) {
		//LOG.trace(TT+"showEditorAsDialog(entity=" + LOG.traceEntity(entity)+(bt==null?"":", upon bt") + ")");
		//System.gc();
		JDialog dlg = new JDialog((Frame) this, true);
		EditorModless = getPref().editorGetModless();
		if (EditorModless) {
			dlg.setModalityType(Dialog.ModalityType.MODELESS);
			dlg.setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
		}
		Editor editor = new Editor(this, entity, dlg);
		dlg.add(editor);
		dlg.setTitle(I18N.getMsg("editor"));
		dlg.pack();
		Dimension pos = SwingUtil.getDlgPosition(entity);
		Dimension size = SwingUtil.getDlgSize(entity);
		if (size != null) {
			dlg.setSize(size.height, size.width);
		} else {
			dlg.setSize(this.getWidth() / 2, 680);
		}
		if (pos != null) {
			dlg.setLocation(pos.height, pos.width);
		} else {
			dlg.setLocationRelativeTo(this);
		}
		if (bt != null && bt.length > 0) {
			Dimension btSize = bt[0].getSize();
			Point loc = dlg.getLocation();
			loc.x += btSize.height;
			loc.y += btSize.width;
			dlg.setLocation(loc);
		}
		dlg.setVisible(true);
		TempUtil.remove(this, entity, Const.FORCE);
		SwingUtil.saveDlgPosition(dlg, entity);
		updated = !editor.canceled;
		return editor.canceled;
	}

	public void backupNew(boolean isManual) {
		String backupDir = getPref().backupGetDir();
		//if backup_directory is empty ask for a directory
		if (isManual && backupDir.isEmpty()) {
			File b = IOUtil.directorySelect(null, "");
			if (b == null) {
				return;
			}
			backupDir = b.getAbsolutePath();
		}
		if (backupDir.isEmpty()) {
			LOG.err("no backup directory, backup failed or canceled");
			return;
		}
		String backupFilename = backupDir + File.separator + h2File.getName();
		//construct backup_filename with properties backup_increment
		if (getPref().backupGetIncrement()) {
			backupFilename += EnvUtil.getDateString();
		}
		File file = new File(backupFilename + ".backup");
		if (file.exists() && isManual
				&& JOptionPane.showConfirmDialog(this,
						I18N.getMsg("file.backup_askreplace", file.getAbsolutePath()),
						I18N.getMsg("backup"),
						JOptionPane.OK_CANCEL_OPTION) == JOptionPane.CANCEL_OPTION) {
			return;
		}
		h2File.doBackup(backupFilename, getPref().backupGetIncrement());
	}

	/**
	 * restor backup file
	 */
	public void backupRest() {
		File b = IOUtil.fileSelect(null, "", "backup", "Backup file (*.backup)", "file.select");
		if (b == null) {
			return;
		}
		//select the directory to restore to
		if (h2File == null) {
			File dir = IOUtil.directorySelect(null, "");
			if (dir == null) {
				return;
			}
			close(false);
			h2File.doRestore(dir, b);
		} else {
			close(false);
			h2File.doRestore(b);
		}
		App.getInstance().openFile(h2File, true);
	}

	public void newEntity(AbstractEntity entity) {
		//LOG.trace(TT+"newEntity(" + entity.getClass().getName() + ")");
		showEditorAsDialog(entity);
	}

	public void windowSaveLayoutAction() {
		String name;
		while (true) {
			name = JOptionPane.showInputDialog(this,
					I18N.getColonMsg("enter.name"),
					I18N.getMsg("docking.save.layout"),
					JOptionPane.PLAIN_MESSAGE);
			File f = new File(EnvUtil.getPrefDir().getAbsolutePath()
					+ File.separator + name + ".layout");
			if (f.exists()) {
				//signaler que le layout existe déjà et demander le remplacement, ou si changer de nom, ou abandonner
				int ret = JOptionPane.showConfirmDialog(this,
						I18N.getMsg("warning") + ":" + I18N.getMsg("docking.save.layout.exists"),
						I18N.getMsg("docking.layout"),
						JOptionPane.YES_NO_OPTION);
				switch (ret) {
					case JOptionPane.OK_OPTION:
						break;
					case JOptionPane.NO_OPTION:
						name = "";
						break;
					default:
						return;
				}
				if (!name.isEmpty()) {
					break;
				}
			} else {
				break;
			}
		}
		DockUtil.layoutSave(this, name);
	}

	public String getClipboardContents() {
		String result = "";
		Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
		//odd: the Object param of getContents is not currently used
		Transferable contents = clipboard.getContents(null);
		boolean hasTransferableText
				= (contents != null)
				&& contents.isDataFlavorSupported(DataFlavor.stringFlavor);
		if (hasTransferableText) {
			try {
				result = (String) contents.getTransferData(DataFlavor.stringFlavor);
			} catch (UnsupportedFlavorException | IOException ex) {
				LOG.err("getClipboard error", ex);
			}
		}
		return result;
	}

}
