/*
 * Copyright (C) 2022 favdb
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package storybook.ui.dialog.preferences;

import i18n.I18N;
import java.awt.event.ActionEvent;
import java.beans.PropertyChangeEvent;
import java.io.File;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;
import javax.swing.border.Border;
import org.miginfocom.swing.MigLayout;
import resources.icons.ICONS;
import resources.icons.IconUtil;
import storybook.App;
import storybook.Pref;
import storybook.toolkit.file.EnvUtil;
import storybook.toolkit.swing.SwingUtil;
import storybook.ui.MIG;
import storybook.ui.panel.AbstractPanel;

/**
 *
 * @author favdb
 */
public class PrefAssistant extends AbstractPanel {

	private static final String TT = "PrefAssistant";
	private static final String ASSISTANT = "assistant.",
			VOGLER = "stage",
			FLAKES = "step",
			POLTI = "polti",
			S_INTERNAL = "internal",
			S_INSTALLED = "installed";

	private final PreferencesDlg caller;
	private Pref pref;
	private JPanel assistantFile;
	private JTextField tfAssistant;
	private Border tfBorder;

	public PrefAssistant(PreferencesDlg dlg) {
		super(dlg.getMainFrame());
		this.caller = dlg;
		initAll();
	}

	@Override
	public void init() {
		pref = mainFrame.getPref();
	}

	@Override
	public void initUi() {
		//LOG.trace(TT + ".initUi()");
		setLayout(new MigLayout(MIG.get(MIG.HIDEMODE3, MIG.WRAP, MIG.INS0, MIG.GAP1), "[][]"));
		String s = pref.getString(Pref.KEY.ASSISTANT).trim();
		if (s.isEmpty()) {
			s = S_INTERNAL;
		}
		ButtonGroup bg = new ButtonGroup();
		JRadioButton rb1 = new JRadioButton(I18N.getMsg(ASSISTANT + S_INTERNAL));
		rb1.setToolTipText(I18N.getMsg(ASSISTANT + S_INTERNAL + "_prompt"));
		bg.add(rb1);
		add(rb1, MIG.get(MIG.SPAN, MIG.split(3)));
		JRadioButton rb2 = new JRadioButton(I18N.getMsg(ASSISTANT + S_INSTALLED));
		rb2.setToolTipText(I18N.getMsg(ASSISTANT + S_INSTALLED + "_prompt"));
		bg.add(rb2);
		add(rb2);
		JRadioButton rb3 = new JRadioButton(I18N.getMsg(ASSISTANT + "external"));
		rb3.setToolTipText(I18N.getMsg(ASSISTANT + "external_prompt"));
		bg.add(rb3);
		add(rb3);
		rb1.addChangeListener(e -> changeAssistant(rb1, rb2, rb3));
		rb2.addChangeListener(e -> changeAssistant(rb1, rb2, rb3));
		rb3.addChangeListener(e -> changeAssistant(rb1, rb2, rb3));
		assistantFile = new JPanel(new MigLayout(MIG.get(MIG.FILLX, MIG.INS0, MIG.GAP0)));
		tfAssistant = new JTextField();
		tfBorder = tfAssistant.getBorder();
		tfAssistant.setColumns(16);
		tfAssistant.setText(s);
		tfAssistant.setCaretPosition(0);
		tfAssistant.addCaretListener(e -> {
			String fn = tfAssistant.getText().trim();
			tfAssistant.setBorder(tfBorder);
			if (S_INTERNAL.equalsIgnoreCase(fn) || S_INSTALLED.equalsIgnoreCase(fn)) {
				return;
			}
			File f = new File(tfAssistant.getText());
			if (!f.exists()) {
				tfAssistant.setBorder(SwingUtil.getBorderRed());
			}
		});
		assistantFile.add(tfAssistant, MIG.get(MIG.GROWX));
		JButton btAssistant = new JButton();
		btAssistant.setIcon(IconUtil.getIconSmall(ICONS.K.F_OPEN));
		SwingUtil.setForcedSize(btAssistant, IconUtil.getDefDim());
		btAssistant.addActionListener((ActionEvent evt) -> {
			String str = tfAssistant.getText();
			if (str.isEmpty()) {
				str = EnvUtil.getPrefDir().getAbsolutePath();
			}
			JFileChooser chooser = new JFileChooser(str);
			chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
			int i = chooser.showOpenDialog(null);
			if (i != 0) {
				return;
			}
			File file = chooser.getSelectedFile();
			tfAssistant.setText(file.getAbsolutePath());
			tfAssistant.setCaretPosition(0);
		});

		assistantFile.add(btAssistant);
		switch (s) {
			case S_INTERNAL:
				tfAssistant.setText(S_INTERNAL);
				assistantFile.setVisible(false);
				rb1.setSelected(true);
				break;
			case S_INSTALLED:
				tfAssistant.setText(S_INSTALLED);
				assistantFile.setVisible(false);
				rb2.setSelected(true);
				break;
			default:
				rb3.setSelected(true);
				break;
		}
		add(assistantFile, MIG.get(MIG.SKIP, MIG.SPAN, MIG.GROWX));
	}

	private void changeAssistant(JRadioButton rb1, JRadioButton rb2, JRadioButton rb3) {
		if (rb1.isSelected()) {
			tfAssistant.setText(S_INTERNAL);
		}
		if (rb2.isSelected()) {
			tfAssistant.setText(S_INSTALLED);
		}
		String tf = tfAssistant.getText();
		if (rb3.isSelected()
				&& (S_INTERNAL.equalsIgnoreCase(tf) || S_INSTALLED.equalsIgnoreCase(tf))) {
			tfAssistant.setText("");

		}
		assistantFile.setVisible(rb3.isSelected());
		caller.pack();
	}

	public void apply() {
		//assistant
		String newAssistant = tfAssistant.getText();
		if (newAssistant.isEmpty()) {
			newAssistant = S_INTERNAL;
		}
		pref.setString(Pref.KEY.ASSISTANT, newAssistant);
		App.getInstance().setAssistant(newAssistant);
	}

	@Override
	public void modelPropertyChange(PropertyChangeEvent evt) {
		// empty
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		// empty
	}

}
