/*
Storybook: Open Source software for novelists and authors.
Copyright (C) 2008 - 2012 Martin Mustun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package storybook.ui.table;

import java.beans.PropertyChangeEvent;
import java.util.List;
import javax.swing.JToolBar;
import org.hibernate.Session;
import resources.icons.ICONS;
import storybook.App;
import storybook.model.EntityUtil;
import storybook.model.Model;
import storybook.model.book.Book;
import storybook.model.hbn.dao.EndnoteDAOImpl;
import storybook.model.hbn.entity.AbstractEntity;
import storybook.model.hbn.entity.DB_DATA;
import storybook.model.hbn.entity.Endnote;
import storybook.ui.MainFrame;
import storybook.ui.Ui;
import static storybook.ui.table.AbsColumn.*;

/**
 * @author martin
 *
 */
@SuppressWarnings("serial")
public class EndnoteTable extends AbstractTable {

	private static final String TT = "EndnoteTable", BT_ENDNOTES = "BtENDNOTES";

	public EndnoteTable(MainFrame mainFrame) {
		super(mainFrame, Book.TYPE.ENDNOTE);
	}

	@Override
	public void init() {
		this.withPart = false;
		allowMultiDelete = false;
	}

	/**
	 * initialize the footer
	 *
	 * @return
	 */
	@Override
	public JToolBar initFooter() {
		JToolBar footer = super.initFooter();
		btNew.setVisible(false);
		btCopy.setVisible(false);
		btDelete.setVisible(false);
		footer.add(Ui.initButton(BT_ENDNOTES, "", ICONS.K.REFRESH,
				"endnotes.renumber", e -> {
					if (Endnote.renumber(mainFrame)) {
						mainFrame.setUpdated();
						fillTable();
					}
				}));
		return footer;
	}

	@Override
	protected void modelPropertyChangeLocal(PropertyChangeEvent evt) {
		// no specific change
	}

	@Override
	protected AbstractEntity getEntity(Long id) {
		Model model = mainFrame.getBookModel();
		Session session = model.beginTransaction();
		EndnoteDAOImpl dao = new EndnoteDAOImpl(session);
		Endnote endnote = dao.find(id);
		model.commit();
		return endnote;
	}

	@Override
	public void updateRow(AbstractEntity entity) {
		if (Book.getTYPE(entity) == Book.TYPE.SCENE) {
			fillTable();
		}
	}

	@Override
	@SuppressWarnings("unchecked")
	public List<AbstractEntity> getAllEntities() {
		List<AbstractEntity> ret = EntityUtil.findEntities(mainFrame, getType());
		return (ret);
	}

	@Override
	public List<AbsColumn> getColumns(AbstractEntity entity) {
		List<AbsColumn> cols = super.getColumns(entity);

		cols.add(new AbsColumn(mainFrame, cols, "endnote.type"));
		cols.add(new AbsColumn(mainFrame, cols, DB_DATA.DATA.NUMBER, NUMERIC, AL_CENTER));
		cols.add(new AbsColumn(mainFrame, cols, DB_DATA.DATA.SCENE, TCR_ENTITY));
		if (App.isDev()) {
			cols.add(new AbsColumn(mainFrame, cols, DB_DATA.DATA.ENDNOTE_SORT));
		}

		this.getColumnsEnd(cols, entity);
		return (cols);
	}

	@Override
	public List<Object> getRow(AbstractEntity entity) {
		//LOG.trace(TT + ".getRow(entity)");
		List<Object> cols = super.getRow(entity);
		Endnote e = (Endnote) entity;
		cols.add(e.getTypeLib());
		cols.add(e.getNumber());
		cols.add(e.getScene());
		if (App.isDev()) {
			cols.add(e.getSort());
		}
		getRowEnd(cols, entity);
		return (cols);
	}

}
